/*
 * Stonage: a puzzle game
 * Copyright (C) 2001 Luc Saillard <luc@saillard.org>
 * 
 * Source for loading, manage Level file
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

#include "stonage.h"
#include "level.h"
#include "image.h"
#include "game.h"
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <SDL.h>
#include <string.h>

/* Please keep in sync with enum sprite_index */
struct enum_list level_sprites[]= {
    {"images/dinosaur.png",0},
    {"images/kinesist.png",0},
    {"images/kinesist2.png",0},
    {"images/key_red.png",KEY_RED},
    {"images/key_blue.png",KEY_BLUE},
    {"images/key_green.png",KEY_GREEN},
    {"images/need_key_red.png",NEED_KEY_RED},
    {"images/need_key_blue.png",NEED_KEY_BLUE},
    {"images/need_key_green.png",NEED_KEY_GREEN},
    {"images/hourglass.png",HOURGLASS},
    {NULL,0}
};

#ifdef ENABLE_DEBUG_TRACE
struct enum_list movement_type_text2enum[] = 
{
   {"move down",MOVE_DOWN},
   {"move up",MOVE_UP},
   {"move right",MOVE_RIGHT},
   {"move left",MOVE_LEFT}
};
#endif




/**
 * Free all memory use by this pointer
 */
void
Level_Free(struct t_level *l)
{
  int i;

  for (i=0;i<l->nsquares;i++)
   {
     Square_Free(&l->squares[i]);
   }
  free(l->squares);

  if (l->map)
   {
     free(l->map);
     l->map=NULL;
   }
  if (l->background)
   {
     SDL_FreeSurface(l->background);
     l->background=NULL;
   }
  free(l);
}

/**
 *
 * Do some to test to check if this level is valid
 * 
 */
int
Level_Sanity_Check(struct t_level *l)
{
  int find_exit;
  int i;
  int *map;

  /* Check that square[x][y]<nsquares and if one exit exist */
  map = l->map;
  for (find_exit=0,i=0; i < (l->sizex*l->sizey);i++,map++)
   {
     if (*map >= l->nsquares)
      {
	trace("x=%d y=%d val=%d",i%l->sizex,i/l->sizex,*map);
	fprintf(stderr,"This level doesn't pass the sanity check.\nAborting\n");
	return 0;
      }
     else if (l->squares[*map].properties&FINISH)
      {
	find_exit=1;
      }
   }

  if (l->start_x==-1 || l->start_y==-1)
   {
     fprintf(stderr,"This level doesn't have a starting point.\nAborting\n");
     return 0;
   }

  if (!find_exit)
   {
     fprintf(stderr,"This level doesn't have at least a ending point.\nAborting\n");
     return 0;
   }

  return 1;
}


/**
 *
 * Draw to this surface the playfield
 *
 */
int 
Level_Draw2Screen(struct t_level *l, SDL_Surface *screen)
{
  int i,x,y,square_index,xbegin,properties;
  SDL_Rect src,dest;
  SDL_Surface *image;
  int *map = l->map;

  /* Compute some values */

  /* All squares have the same size (for now) */
  dest.w = l->squares[1].images[0]->w;
  dest.h = l->squares[1].images[0]->h;

  xbegin = (screen->w - (dest.w * l->sizex))/2;
  dest.y = (screen->h - (dest.h * l->sizey))/2;
  
  /* Draw the background */
  if (l->background==NULL)
   {
     /* Fill the backcolor in black */
     SDL_FillRect(screen,NULL,0);
   }
  else
   {
     SDL_BlitSurface(l->background, NULL, screen, NULL);
   }
  /* Draw the playfield */
  for (y=0;y<l->sizey;y++)
   {
     dest.x = xbegin;
     for (x=0;x<l->sizex;x++)
      {
	square_index = *map++;
	if (square_index)
	 {
	   /* Draw the square */
	   image = l->squares[square_index].images[0];
	   SDL_BlitSurface(image, NULL, screen, &dest);
	   properties = l->squares[square_index].properties;
	   /* Draw the sprite */
	   for (i=0;level_sprites[i].text;i++)
	    {
	      if (properties & level_sprites[i].flags)
	       {
		 SDL_BlitSurface(l->squares[0].images[i], NULL, screen, &dest);
		 break; /* for now, we have only one capability sprite */
	       }
	    }
	 }
	dest.x +=dest.w;
      }
     dest.y+=dest.h;
   }

  /* Draw the dinosaur */
  dest.x = xbegin + (dest.w * l->xdino);
  dest.y = ((screen->h - (dest.h * l->sizey))/2) + (dest.h * l->ydino);
  SDL_BlitSurface(l->squares[0].images[SPRITE_DINOSAUR], NULL, screen, &dest);

  /* ... and draw the kinesist sprite */
  if (l->control==CONTROL_KINESIST ||
      l->control==CONTROL_SQUARE)
   {
     dest.x = xbegin + (dest.w * l->xkinesist);
     dest.y = ((screen->h - (dest.h * l->sizey))/2) + (dest.h * l->ykinesist);
     SDL_BlitSurface(l->squares[0].images[l->control==CONTROL_KINESIST?SPRITE_KINESIST_NOTSELECTED:SPRITE_KINESIST_SELECTED], NULL, screen, &dest);
   }

  /* ... and draw the timeline */
  dest.x=0; 
  dest.y=screen->h-16;
  src.x=0;
  src.y=0;
  src.w=(game.time_remaining * l->timebar->w / l->times[game.difficulty]);
  src.h=8;
  SDL_BlitSurface(l->timebar,&src, screen, &dest);

  return 1;
}


/*
 * Function: Level_TestCanGoAbove
 * Comment:
 *   We test if the dinausor can go abose the square, we make a function,
 *   because we can test if the user have the good key, and all other actions.
 * INPUT:
 * 
 * OUTPUT:
 *  0: can't move
 *  1: ok.
 */

static int
Level_TestCanGoAbove(struct t_level *l, unsigned int square_properties)
{
  if (   (square_properties&CAN_GO_ABOVE)
      && !((square_properties&NEED_KEY_RED) && !(l->keys&KEY_RED))
      && !((square_properties&NEED_KEY_BLUE) && !(l->keys&KEY_BLUE))
      && !((square_properties&NEED_KEY_GREEN) && !(l->keys&KEY_GREEN)))
   {
     return 1;
   }
  return 0;
}

/*
 * Function: Level_MoveDinoTo
 * INPUT:
 *  direction: 0 down, 1:up, 2:right, 3:left
 *  automove: true,false
 * OUTPUT:
 *  0: can't move
 *  1: The dino have moved. (and optionnaly the square)
 *  2: The dino AND the square have moved.
 */
int
Level_MoveDinoTo(struct t_level *l, enum movement_type direction, int automove)
{
  unsigned int properties;
  int *map;

  map = l->map + (l->ydino * l->sizex) + l->xdino;
  properties = l->squares[*map].properties;

  trace ("direction: %s",movement_type_text2enum[direction].text);

  switch(direction)
   {
     /* Each test is done in that order:
      *   1. Test against border of the playfield
      *   2. Test if the dinosaur can move to the next square
      *   3. Test if the dinosaur is on a moveable square and the square can
      *      move in the good direction.
      *    3-1. Test if the square can move (a empty case)
      *         Then move the square and the dinosaur
      */
    case MOVE_RIGHT:
      if (l->xdino >= l->sizex-1)
	break;
      map++;
      if (!automove && Level_TestCanGoAbove(l,l->squares[*map].properties))
       {
	 l->xdino++;
	 return 1;
       }
      if (properties&CAN_MOVE_TO_EAST)
       {
	 if (*map==0)
	  {
	    l->xdino++;
	    *map=*(map-1);
	    *(map-1)=0;
	    return 2;
	  }
       }
      break;

    case MOVE_LEFT:
      if (l->xdino <= 0)
	break;
      map--;
      if (!automove && Level_TestCanGoAbove(l,l->squares[*map].properties))
       {
	 l->xdino--;
	 return 1;
       }
      if (properties&CAN_MOVE_TO_WEST)
       {
	 if (*map==0)
	  {
	    l->xdino--;
	    *map=*(map+1);
	    *(map+1)=0;
	    return 2;
	  }
       }
      break;

    case MOVE_UP:
      if (l->ydino < 1)
	break;
      map-=l->sizex;
      if (!automove && Level_TestCanGoAbove(l,l->squares[*map].properties))
       {
	 l->ydino--;
	 return 1;
       }
      if (properties&CAN_MOVE_TO_NORTH)
       {
	 if (*map==0)
	  {
	    l->ydino--;
	    *map=*(map + l->sizex);
	    *(map + l->sizex)=0;
	    return 2;
	  }
       }
      break;

    case MOVE_DOWN:
      if (l->ydino >= l->sizey-1)
	break;
      map+=l->sizex;
      if (!automove && Level_TestCanGoAbove(l,l->squares[*map].properties))
       {
	 l->ydino++;
	 return 1;
       }
      if (properties&CAN_MOVE_TO_SOUTH)
       {
	 if (*map==0)
	  {
	    l->ydino++;
	    *map=*(map - l->sizex);
	    *(map - l->sizex)=0;
	    return 2;
	  }
       }

      break;

    default:
      return 0;
   }
  /* Special case for teleport square */
  if (properties&(TELEPORT1|TELEPORT2|TELEPORT3))
     return 1;
  return 0;
}

/*
 * Function: Level_MoveKinesistTo
 * INPUT:
 * direction: 0 down, 1:up, 2:right, 3:left
 * OUTPUT:
 *  0: can't move
 *  1: ok.
 */
int
Level_MoveKinesistTo(struct t_level *l, enum movement_type direction)
{
  int *map;
  /* Some sanity check over buffer overflow */
  if (  (direction==MOVE_RIGHT && l->xkinesist >= l->sizex-1)
      ||(direction==MOVE_LEFT  && l->xkinesist == 0)
      ||(direction==MOVE_DOWN  && l->ykinesist >= l->sizey-1)
      ||(direction==MOVE_UP    && l->ykinesist == 0))
   {
     return 0;
   }
  
  map = l->map + (l->ykinesist * l->sizex) + l->xkinesist;
  trace("direction=%d",direction);
  switch(direction)
   {
    case MOVE_RIGHT:
      l->xkinesist++;
      return 1;

    case MOVE_LEFT:
      l->xkinesist--;
      return 1;

    case MOVE_UP:
      l->ykinesist--;
      return 1;

    case MOVE_DOWN:
      l->ykinesist++;
      return 1;

    default:
      return 0;
   }
}

/*
 * Function: Level_CanMoveSquare
 * Comment:
 *   Return all direction where this square can go. If the square can't move,
 *   return 0; Direction is the same as properties
 * Input:
 *   l: the level
 * Output:
 *   0|CAN_MOVE_TO_SOUTH|CAN_MOVE_TO_NORTH|CAN_MOVE_TO_WEST|CAN_MOVE_TO_EAST
 */
int
Level_CanMoveSquare(const struct t_level *l)
{
  int square_can_move;
  unsigned int properties;
  int *map;

  map = l->map + (l->ykinesist * l->sizex) + l->xkinesist;
  properties = l->squares[*map].properties;

  /* If the dinosaur is on the square, this square can't move this square */
  if ((l->xkinesist == l->xdino) && (l->ykinesist == l->ydino))
    return 0;

  square_can_move=0;

  properties&=CAN_MOVE_TO_MASK;
  if (!properties) /* If the square can't move, exit now */
    return 0;

  /* For each direction,
   *  test if the square can move to this place,
   */
  if (properties&CAN_MOVE_TO_WEST)
   {
     if (*(map - 1)==0)
      {
	square_can_move|=CAN_MOVE_TO_WEST;
      }
   }
  
  if (properties&CAN_MOVE_TO_EAST)
   {
     if (*(map + 1)==0)
      {
	square_can_move|=CAN_MOVE_TO_EAST;
      }
   }

  if (properties&CAN_MOVE_TO_NORTH)
   {
     if (*(map - l->sizex)==0)
      {
	square_can_move|=CAN_MOVE_TO_NORTH;
      }
   }
  
  if (properties&CAN_MOVE_TO_SOUTH)
   {
     if (*(map + l->sizex)==0)
      {
	square_can_move|=CAN_MOVE_TO_SOUTH;
      }
   }

#if ENABLE_DEBUG_TRACE
   { int i;
     for (i=0;i<32 && square_properties_text2enum[i].text;i++)
      {
	if (square_properties_text2enum[i].flags & square_can_move)
	  trace("square_can_move: %d (%s)",square_can_move,square_properties_text2enum[i].text);
      }
   }
#endif
  
  return square_can_move;
}

/*
 * Function: Level_MoveSquare
 * Comment:
 *   Move the square in the direction indicated
 * Input:
 *
 * Output:
 *   return TRUE if the square have moved.
 */
int
Level_MoveSquareTo(struct t_level *l, enum movement_type direction)
{
  int properties;
  int *map;

  map = l->map + (l->ykinesist * l->sizex) + l->xkinesist;
  properties = l->squares[*map].properties;

  trace ("direction: %s",movement_type_text2enum[direction].text);
  if (direction == MOVE_RIGHT && (properties&CAN_MOVE_TO_EAST))
   {
     map++;
     if ((l->xkinesist < l->sizex-1) && *map==0)
      {
	l->xkinesist++;
	*map=*(map-1);
	*(map-1)=0;
	return 1;
      }
     return 0;
   }
  else if (direction == MOVE_LEFT && (properties&CAN_MOVE_TO_WEST))
   {
     map--;
     if ((l->xkinesist > 0) && *map==0)
      {
	l->xkinesist--;
	*map=*(map+1);
	*(map+1)=0;
	return 1;
      }
     return 0;
   }
  else if (direction == MOVE_UP && (properties&CAN_MOVE_TO_NORTH))
   {
     map-=l->sizex;
     if ((l->ykinesist > 0) && *map==0)
      {
	l->ykinesist--;
	*map=*(map + l->sizex);
	*(map + l->sizex)=0;
	return 1;
      }
     return 0;
   }
  else if (direction == MOVE_DOWN && (properties&CAN_MOVE_TO_SOUTH))
   {
     map+=l->sizex;
     if ((l->ykinesist < l->sizey-1) && *map==0)
      {
	l->ykinesist++;
	*map=*(map - l->sizex);
	*(map - l->sizex)=0;
	return 1;
      }
     return 0;
   }
 return 0;
}

/**
 *
 * Find the next teleporter in the level
 *
 */
static int
Level_FindTeleporter(struct t_level *l, int teleportnumber)
{
  int *map = l->map;
  int x,y,square_index;
  for (y=0;y<l->sizey;y++)
   {
     for (x=0;x<l->sizex;x++)
      {
	square_index = *map++;
	if (square_index)
	 {
	   if ((l->squares[square_index].properties&teleportnumber) &&
	       x != l->xdino && y != l->ydino)
	    {
	      /* Warp the dino to this position */
	      l->xdino = x;
	      l->ydino = y;
	      return 1;
	    }
	 }
      }
   }
  return 0;
}


/**
 * Function: Level_DoPostMoveAction
 * Comment:
 *   Execute function after the move and before the animation.
 * Input:
 *
 * Output:
 *   TRUE or FALSE
 */
int
Level_DoPostMoveAction(struct t_level *l)
{
  int square_index;
  unsigned int properties, *map;
  
  map = l->map + (l->ydino * l->sizex) + l->xdino;
  square_index=*map;
  properties = l->squares[square_index].properties;

  if ((properties&KEY_RED) || (properties&NEED_KEY_RED))
   {
     l->keys|=KEY_RED;
     properties&=~(KEY_RED|NEED_KEY_RED);
   }
  if ((properties&KEY_GREEN) || (properties&NEED_KEY_GREEN))
   {
     l->keys|=KEY_GREEN;
     properties&=~(KEY_GREEN|NEED_KEY_GREEN);
   }
  if ((properties&KEY_BLUE) || (properties&NEED_KEY_BLUE))
   {
     //*map = Square_FindProperties(l->squares,l->nsquares,properties);
     //if (*map == -1)
     //  *map = 0; /* It's a bug ... */
     l->keys|=KEY_BLUE;
     properties&=~(KEY_BLUE|NEED_KEY_BLUE);
   }

  if (properties&TELEPORT1)
     Level_FindTeleporter(l,TELEPORT1);
  else if (properties&TELEPORT2)
     Level_FindTeleporter(l,TELEPORT2);
  else if (properties&TELEPORT3)
     Level_FindTeleporter(l,TELEPORT3);

  if (properties&HOURGLASS)
   {
     l->times[game.difficulty]*=2;
     *map = Square_FindProperties(l->squares,l->nsquares,properties&~HOURGLASS);
     if (*map == -1)
       *map = 0; /* It's a bug ..., so draw it like a bug */
   }

  /**********************************/
  /* Do action on the last position */
  /**********************************/

  map = l->map + (l->ydino_old * l->sizex) + l->xdino_old;
  properties = l->squares[*map].properties;

  if (properties&AFTER_MOVE_CRASH)
   {
     *map=0;
   }

  /* Save the last position */
  l->xdino_old=l->xdino;
  l->ydino_old=l->ydino;

  return 0;
}


/*
 * Function: Level_IsFinish
 * Comment:
 *   return true when the user have accomplished all tasks. For now,
 *   we only test if the user is on a Finish square
 * Input:
 *
 * Output:
 *   TRUE or FALSE
 */
int
Level_IsFinish(struct t_level *l)
{
  int square_index;
  unsigned int properties;
  
  square_index=l->map[ (l->ydino * l->sizex) + l->xdino];
  properties = l->squares[square_index].properties;

  if (properties&FINISH)
    return 1;
  else
    return 0;
}


